# gsa-is

This code is an implementation of gSAIS and gSACA-K \[1\], which
extend the linear-time suffix sorting algorithms SAIS \[2\] and SACA-K \[3\] to
compute the generalized suffix array, maintaining their theoretical bounds,
respecting the order among all suffixes, and improving their practical
performance.

Overall, gSACA-K's time-space trade-off is Pareto optimal compared to the all other algorithms in the experiments.


## run (SA and LCP) 

To run a test type:

```sh
make
./gsais dataset/input-10000.txt 10000 -m 6 -l -o
```

This command should produce a file with extension *.sa_lcp* 
containing the sa and lcp values interleaved 
(4+4 bytes little endian format). If *-l* is omitted  only sa values are stored in a file with extension *.sa*. 

One can change to 64 bits integers (when n >= 2^31) in lib/utils.h, setting m64 to 1.


## run (BWT and LCP)

To run a BWT test type:

```sh
make
./gsais long_text.txt 10000 -m 6 -b -g 2
```
This will produce a file with extension *.bwt* with the BWT of the collection and a file with extension .2.lcp with the LCP values stored 
in 2 bytes unsigned integers little endian format. The writing fails if one LCP value does not fit in 2 bytes (ie is larger than 65535). Legal values for the *-g* options are between 1 and 7. If the *-g* option is missing only the bwt values are computed and saved to file 


## input format

The program recognizes files with extension .txt (one document per line), .fastq (one document per read), .fasta (one document per sequence). See file *lib/file.c* for details.

Another possibility is to use the plain *raw + len* format using the option *-L*. In this case the input file must contain the concatenation of the input documents, and the file with extension *.len* contains the length of the single documents, in 4-bytes little endian format.

Note that .txt .fastq and .fasta format can be converted to the *raw + len* format using the *-X* option. 


## Settings

-m parameter specifies which algorithm is called by main.c:

* 1: SAIS\* 
* 2: SACA-K\* 
* 3: SAIS  
* 4: SACA-K 
* 5: gSAIS
* 6: gSACA-K

SAIS\* and SACA-K\* are versions that receive an integer alphabet as input.



## Validate

One can check if the output produced is correct with the option -c


## Output:

One can output the GSA produced as .sa with the option -o
If the option -l is given also lcp is computed and lcp and sa are written in a single .sa_lcp file

## References

\[1\] Louza, F. A., Gog, S., Telles, G. P., Induced Suffix Sorting for String Collections. In Proc. DCC, pp. 43-58, 2016, [Preprint](http://www.ic.unicamp.br/~ra144643/downloads/articles/2016-dcc.pdf), [bibtex](http://www.ic.unicamp.br/~ra144643/downloads/bib/2016-dcc.bib).

\[2\] Nong G., Zhang S., Chan W. H., Two efficient algorithms for linear time suffix array construction, IEEE Trans. Comput., vol. 60, no. 10, pp. 1471–1484, 2011

\[3\] Nong, G., Practical linear-time O(1)-workspace suffix sorting for constant alphabets, ACM Trans. Inform. Syst., vol. 31, no. 3, pp. 1–15, 2013


## Final note

Most of the material, in particular all the code for the construction of SA and LCP arrays has been downloaded from the repository [https://github.com/felipelouza/gsa-is](https://github.com/felipelouza/gsa-is) on October 2016. 
GM has changed *main.c* to support command line options and the *raw+len* format.

