

/**
 * Using the number of occs of each symbol in each bwt (stored in bwtOcc)
 * init the array Z (mergeColor) and B (blockBeginsAt) at the value
 * they should have after the first iteration of the Gap algorithm:
 *   blockBeginsAt[i]=1 if i is the first occurrence in the first
 *                      column F of a new symbol or F[i]=0
 *                      (0 occurrences are assumed to be all different)
 *   in each region of the F column with the same symbol j in Z
 *   we have: #occ(j) in bwt[0], #occ(j) in bwt(1), and so on
 * Since the region corresponding to 0 does not chage and 0 has a 
 * special udate rule, we init that region also in Znew (newMergeColor)
 * and we never modify thet region in the algorithm.   
 * The array firstColumn (compact representation of F) is also initialized
 *  
 * */     
// init Z, newZ B, and first Column array using g->bwtOcc[i][j]
static void init_arrays8(g_data *g)
{
  assert(g->numBwt<=8);
  customInt i=0; // position inside Z newZ and B  

  for(int j=0;j<g->sizeOfAlpha;j++) {
    tba_or_m8(g->mergeColor,i,1);
    g->firstColumn[j] = i;  // symbol j starts at position i
    for(int b=0;b<g->numBwt;b++) {
      for(customInt t=0;t<g->bwtOcc[b][j];t++) {
        if(ktuple_tail0(j))   // zero chars are all different, Z are newZ do not change 
          tba_or_m8(g->mergeColor,i,1);
        g->mergeColor[i++] |= b*9;      // write b in both cur and next
      } // end for t
    } // end for b 
  } // end for j
  assert(i==g->mergeLen); 
  // extra check on mergeColor, can be commented out
  #ifndef NDEBUG
  customInt cnt[MAX_NUMBER_OF_BWTS] = {0};
  for(i=0;i<g->mergeLen;i++) cnt[g->mergeColor[i]&0x7]++;
  bool stop=false;
  for(int i=0; i<g->numBwt; i++)
    if(cnt[i]!=g->bwtLen[i]) {
      printf("INIT %d cnt:"CUSTOM_FORMAT" len:"CUSTOM_FORMAT"\n",i,cnt[i],g->bwtLen[i]);
      stop=true;
    }
  assert(!stop);
  #endif
}


// init Z, newZ and B array without using g->bwtOcc[i][j]
static void init_arrays8_largealpha(g_data *g)
{
  // compute first column and a copy of it for the chars in bwt[0]
  customInt t,tot, cnt[SIZE_OF_ALPHABET];
  ktuple_init_freq(g->bws[0],g->mergeLen,cnt); 
  int j;
  tot=0;
  for(j=0;j<g->sizeOfAlpha;j++) {
    t = cnt[j];
    g->firstColumn[j] = cnt[j] = tot;
    tot += t;
  }
  assert(tot==g->mergeLen);
  
  // init the B array: mark where LCP is < g->charXsymb
  customInt num,i; 
  for(j=0;j<g->sizeOfAlpha;j++) {
    // count number of occurrences of j
    if(j<g->sizeOfAlpha-1)  num = g->firstColumn[j+1] - g->firstColumn[j];
    else                    num = g->mergeLen - g->firstColumn[j];
    assert(j!=0 || num>0);  // 0 must be present 
    if(num>0) {
      assert(!ktuple_has0(j) || ktuple_tail0(j));// the only ktuple with 0 are tail0 
      i = g->firstColumn[j];
      // mark entry B[i]
      tba_or_m8(g->mergeColor,i,1);
      // mark following entries if tail 0 chars
      if(ktuple_tail0(j)) {
        for(customInt k=1; k<num; k++) {
          tba_or_m8(g->mergeColor,i+k,1);
        }
      }
    }
  } // end j, B array OK
  
  // init Z, newZ by scanning the bwts
  for(int b=0;b<g->numBwt;b++) {
    for(customInt i=0;i<g->bwtLen[b];i++) {
      int j = ktuple_decode0(g->bws[b][i]);
      g->mergeColor[ cnt[j]++ ] |= b*9; // write b in both cur and new 
    }
  }        
  #ifndef NDEBUG
  // extra check on cnt
  for(i=0;i<g->sizeOfAlpha-1;i++) assert(cnt[i]==g->firstColumn[i+1]);
  assert(cnt[g->sizeOfAlpha-1] == g->mergeLen);
  // extra check on mergeColor
  customInt cnx[MAX_NUMBER_OF_BWTS] = {0};
  for(i=0;i<g->mergeLen;i++) cnx[g->mergeColor[i]&0x7]++;
  bool stop=false;
  for(int i=0; i<g->numBwt; i++)
    if(cnx[i]!=g->bwtLen[i]) {
      printf("INIT %d cnx:"CUSTOM_FORMAT" len:"CUSTOM_FORMAT"\n",i,cnx[i],g->bwtLen[i]);
      stop=true;
    }
  assert(!stop);
  #endif
}


// single iteration of the Gap algorithm
// input is head of the irrelevant list
#define get_mergeColor8(k,round) ((round)? (g->mergeColor[(k)]>>3)&0x07 : (g->mergeColor[(k)]&0x07))
#define set_mergeColor8(k,c,round) (g->mergeColor[(k)] = (round)? \
        ( (g->mergeColor[(k)]&0xF8) | (c)) : ((g->mergeColor[(k)]&0xC7) | ((c)<<3))  )


static solidBlock *addCharToPrefix8(solidBlock *ibHead, liquidBlock *liquid, bool *mergeChanged, const int round, g_data *g) {
  assert(liquid->empty);
  liquid->beginsAt = liquid->endsAt = 0;  
  for(int i=0;i<liquid->occ_size;i++) assert(liquid->occ[i]==0);
  // copy first column to F 
  array_copy(g->F, g->firstColumn, g->sizeOfAlpha); //initialize char positions
  // pointer inside each BWT (k_0 & k_1 in the pseudocode)
  array_clear(g->inCnt,g->numBwt,0);
  // id for each character, init with an invalid id
  customInt blockID[g->sizeOfAlpha];
  array_clear(blockID,g->sizeOfAlpha, g->mergeLen);  // mergeLen is an invalid id
  customInt id = 0, k; 
  int m = (round%2==1) ? 1 : 2; // mask for the bitB array (now inside mergeColor)

  protoBlock cblock = {.mono = false};
  solidBlock *next = ibHead;
  solidBlock *last = NULL;
  
  for (k = 0; k < g->mergeLen; ) { 
    assert(next==NULL || k <= next->beginsAt);  // we did not pass next block
    assert(last==NULL || last->nextBlock == next); // last is the immediately preceeding block

    // check if we are entering a block, and if the block is at leat 2 iterations old
    bool start_block, last_block_recent=true; // for k=0 a new block starts, so last id properly initialized
    start_block = tba_block_test_set8(g->mergeColor,k,m,&last_block_recent);    
    if (start_block) {
      // check if the block we just left is not recent and monocrhome 
      if(cblock.mono==true && !last_block_recent) {
        cblock.endsAt = k;           // solidifiable monochorome block just ended
        add_proto2liquid(&cblock,liquid);  // add proto to liquid that remains active 
      }
      else { // proto block cannot be added, close current liquid
        if(!liquid->empty) {
          last = finalize_liquid(last,liquid,next); // this is the only point where a new block is created 
          if(last!=NULL && ibHead==last->nextBlock) ibHead = last;
        }
        assert(liquid->empty);
        liquid->beginsAt=liquid->endsAt=k; // start empty liquid block
      }
      assert(liquid->endsAt==k);
      // block ending at k considered, now look forward 
      if(next!=NULL && next->beginsAt==k) { // entering an irrelevant block
        skip(next, g);                      // skip block 
        k = next->endsAt;                   // update k
        // merge liquid with next block and possibly previous 
        if(last==NULL || last->endsAt!=liquid->beginsAt) {
            if(!liquid->empty) merge_liquid(liquid,next); // simple merge 
            last = next;   // advance last           
        } 
        else  //three way merge, next is freed last does not change
          merge_sls(last,liquid,next); 
        assert(liquid->empty);
        liquid->beginsAt=liquid->endsAt=k; // start empty liquid block
        next = last->nextBlock;            // next has become last, update next 
        assert(k==last->endsAt);
        cblock.mono = false;        // prevent re-adding the just skipped block
        continue;                   // resume from the end of the block 
      }
      // we are entering a relevant block, unless it is a recent one make it a candidate for solidification  
      if( !last_block_recent ) {
        cblock.beginsAt = k; cblock.mono=true; 
        cblock.color =  get_mergeColor8(k,round);     // g->mergeColor[k];
        cblock.start = &g->bws[cblock.color][g->inCnt[cblock.color]]; // bwt-position of first char in block
      }
      else cblock.mono = false; // not a candidate for solid block, wait next iteration 
      if(last_block_recent) 
         id = k;    // id of the new block 
    }
    
    // processing a char in a relevant block
    int currentColor = get_mergeColor8(k,round);   // g->mergeColor[k] b in pseudocode
    int currentChar = g->bws[currentColor][g->inCnt[currentColor]++]; // c in pseudocode
    if(currentColor != cblock.color) cblock.mono = false;       // block is not monochrome
    // write color in new Z array, except 0 chars
    if(!ktuple_has0(currentChar)) {
      customInt positionToUpdate = g->F[currentChar]++;
      set_mergeColor8(positionToUpdate,currentColor,round); // g->newMergeColor[positionToUpdate] = currentColor;
      if(g->bwtOnly && !*mergeChanged && !g->mmapZ && get_mergeColor8(positionToUpdate,round)!=currentColor)
        *mergeChanged=true;  // remember there is a difference from the previous ieration 
      // create new block?
      if (blockID[currentChar] != id) {
        if(last_block_recent) tba_mark8_if0(g->mergeColor,positionToUpdate,m);
        blockID[currentChar] = id; // update block id, always!
      }
    }
    k++;
  } // end main loop
  assert(next==NULL); 
  if(cblock.mono==true) { // final mono block can be created 
    cblock.endsAt = k;
    add_proto2liquid(&cblock,liquid);  // add proto to liquid that remains active
    assert(!liquid->empty);
  }
  if(!liquid->empty) {
    last = finalize_liquid(last,liquid,NULL); // a new block could be created 
    if(last!=NULL && ibHead==last->nextBlock) ibHead = last;
  }
  assert(liquid->empty);
  // check we have read all chars from all BWT's 
  for(int i=0; i<g->numBwt; i++)
    assert(g->inCnt[i]==g->bwtLen[i]);

  // no need to swap merge and newMerge!
  if(g->charXsymb>1 && !g->bwtOnly) { // swap diffChar and newDiffChar
    symbol *tmp=g->diffChar; g->diffChar = g->newDiffChar; g->newDiffChar=tmp;
  }
  return ibHead;
}


// entry point for the gap bwt/lcp merging procedure
// if we are only interested in BWT blockBeginsAt can be replaced by a bit array 
void gap8(g_data *g, bool lastRound) {

  // check we can really use 4 bits
  assert(g->numBwt<=8 && g->bwtOnly);
  if(g->verbose>1) puts("BWT merging with gap8");
  // init local global vars
  check_g_data(g);
    
  // allocate Z (merge) Znew; and clean it since it is also used for B 
  assert(g->blockBeginsAt==NULL);
  alloc_merge_array(g);
  memset(g->mergeColor,0, g->mergeLen*sizeof(palette));
  
  // allocate other useful arrays
  g->inCnt = malloc(g->numBwt*sizeof(customInt)); 
  g->firstColumn = malloc(g->sizeOfAlpha*sizeof(customInt));
  g->F = malloc(g->sizeOfAlpha*sizeof(customInt));
  if(!g->inCnt || !g->firstColumn || !g->F) 
    die(__func__);
    
  // if squeezed and lcp allocate diffChar
  if(g->charXsymb>1 && !g->bwtOnly) {
    g->diffChar= (symbol *) malloc(g->mergeLen*sizeof(symbol));
    g->newDiffChar= (symbol *) malloc(g->mergeLen*sizeof(symbol));
    if(!g->diffChar || !g->newDiffChar) die(__func__);
  }
  // init the above arrays
  if(g->smallAlpha) init_arrays8(g);
  else init_arrays8_largealpha(g);
  
  // list of the irrelevant blocks initially empty
  solidBlock *ibList = NULL;
  customInt prefixLength = 1;
  // init liquid block (containing list of allocated mem)
  liquidBlock *liquid = liquid_new(g); 

  if (g->verbose>2) printf("-----------> Initial values:\n");
  if (g->verbose>2) testPrintMergeColor(prefixLength,g->mergeColor, g->mergeLen);

  // main loop
  int round=0;
  if(g->numBwt>1) {
    while ((ibList == NULL) || (ibList->endsAt < g->mergeLen) || ibList->beginsAt!=0) {
      prefixLength+= g->charXsymb;
      if(prefixLength>MAX_LCP_SIZE && !g->bwtOnly) {fprintf(stderr,"LCP too large\n");die(__func__);}
      bool mergeChanged = false; // the Z vector has changed in this iteration (used when g->bwtOnly)
      ibList=addCharToPrefix8(ibList,liquid,&mergeChanged,round,g);
      assert(checkBlocks(ibList,prefixLength,g,lastRound)==0); //\\ check no blocks are mergeable
      if (g->verbose>2 && lastRound) showIrrelevants(ibList);
      if (g->verbose>1 && lastRound) {
        printf("Lcp: "CUSTOM_FORMAT". Memory: %zu peak, %zu current, %.4lf/%.4lf bytes/symbol\n", 
             prefixLength-1, malloc_count_peak(),
             malloc_count_current(), (double)malloc_count_peak()/g->mergeLen,
             (double)malloc_count_current()/g->mergeLen);
      }
      if(g->bwtOnly && !mergeChanged) {
        if(g->verbose>1) puts("Gap bwt-only early termination");
        break;
      }
      round  = 1 - round; // change round parity
    }
    solidList_free(ibList,liquid);
  }
  if (g->verbose>1) {
    if(lastRound)
      printf("Merge completed (%d bwts). Mem: %zu peak, %zu current, %.2lf/%.2lf bytes/symbol\n", g->numBwt, malloc_count_peak(),
           malloc_count_current(), (double)malloc_count_peak()/g->mergeLen,
           (double)malloc_count_current()/g->mergeLen);
    else if(g->verbose>1)
      printf("Merge completed (%d bwts). Mem: %zu peak, %zu current\n", g->numBwt, malloc_count_peak(),
           malloc_count_current());
  }
  liquid_free(liquid);

  // computation complete, do the merging. The following call writes
  // the merged BWT to g->newMergeColor, the LCP (if present) to g->blockBeginsAt
  // normalize mergeColor
  if(round!=0) for(customInt i=0;i<g->mergeLen;i++) g->mergeColor[i] = (g->mergeColor[i]>>3) & 0x07;
  else for(customInt i=0;i<g->mergeLen;i++) g->mergeColor[i] &= 0x07;
  mergeBWTandLCP(g);
  // copy the merged bwt back to bws[0]
  if(lastRound) { // if last round remap while copying 
    symbol *tmp = (symbol *) g->mergeColor;
    for(customInt i=0;i<g->mergeLen;i++)
      g->bws[0][i] = alpha_enlarge(ktuple_last(tmp[i]));
  }
  else  
    memcpy(g->bws[0],g->mergeColor,g->mergeLen*sizeof(symbol));

  free(g->F); // last five arrays deallocated
  free(g->firstColumn); 
  free(g->inCnt);
  free_merge_array(g);
}

#undef get_mergeColor8
#undef set_mergeColor8
